import os
import typer
from .utils import (
    clear,
    check_internet_connection,
    is_youtube_video_id,
    validate_link,
    handle_video_link,
)
from .downloader import download


# Create CLI app
app = typer.Typer(
    name="pyutube",
    add_completion=False,
    help="Awesome CLI to download YouTube videos(as video or audio)/shorts from the terminal",
    rich_markup_mode="rich",
)


@app.command(
    name="download",
    help="Download a [red]YouTube[/red] videos (as video or audio), shorts, (playlists soon as possible).",
    epilog="Made with ❤️ By Ebraheem. Find me on GitHub: [link=https://github.com/Hetari]click here @Hetari[/link]"
)
def pyutube(
    url: str = typer.Argument(
        ...,
        help="YouTube URL",
        show_default=False),
    path: str = typer.Argument(
        os.getcwd(),
        help="Path to save video [default: <current directory>]",  show_default=False),
):
    """
    Downloads a YouTube video.

    Args:
        url (str): The URL of the YouTube video.
        path (str): The path to save the video. Defaults to the current working directory.

    Returns:
        None
    """
    clear()

    if not check_internet_connection():
        return

    if is_youtube_video_id(url):
        url = f"https://www.youtube.com/watch?v={url}"

    is_valid_link, link_type = validate_link(url)

    if not is_valid_link:
        return

    if link_type == "video":
        try:
            url, path, is_audio = handle_video_link(url, path)
        except TypeError:
            return

        download(url, path, is_audio)

    elif link_type == "short":
        download(url, path, is_audio)



from yaspin import yaspin
from yaspin.spinners import Spinners
from pytube import YouTube
from pytube.cli import on_progress
from termcolor import colored
from .utils import (
    console,
    error_console,
    sanitize_filename,
    ask_rename_file,
    ask_resolution,
    rename_file,
    is_file_exists,
)


class Downloader:
    def __init__(self, url: str, path: str, is_audio: bool = False, is_playlist: bool = False, is_short: bool = False):
        self.url = url
        self.path = path
        self.quality = "720p"
        self.is_audio = is_audio
        self.is_playlist = is_playlist
        self.is_short = is_short
        self.video_id = None

    @yaspin(text=colored("Searching for the video", "green"), color="green", spinner=Spinners.point)
    def video_search(self, url: str) -> YouTube:
        """
        This function searches for a video using the provided URL and returns an
        instance of the YouTube class representing the searched video.

        Args:
            url: The URL of the video to search for.

        Returns:
            YouTube: An instance of the YouTube class representing the searched video.
        """

        return YouTube(
            url,
            use_oauth=False,
            allow_oauth_cache=True,
            on_progress_callback=on_progress,
        )

    @yaspin(text=colored("getting video headers ", "green") + colored("(means the video won't be fully downloaded)", "yellow"), spinner=Spinners.point)
    def get_available_resolutions(self, video: YouTube) -> set:
        """
        Get a set of all available resolutions for the video.

        Args:
            video: The video to retrieve available resolutions from.

        Returns:
            set: A set containing all available resolutions.
        """
        available_streams = video.streams.filter(progressive=True)
        resolutions = {
            stream.resolution for stream in available_streams if stream.resolution
        }
        return resolutions, available_streams

    @yaspin(text=colored("Downloading the video...", "green"), color="green", spinner=Spinners.dots13)
    def get_video_streams(self, video: YouTube, quality: str, streams: YouTube.streams) -> YouTube:
        """
        Downloads the video streams based on the specified quality.
        If the specified quality is not available, it selects the nearest quality.

        Args:
            video: The video to retrieve streams from.
            quality: The desired quality of the video streams.

        Returns:
            The video stream with the specified quality, or the best available stream if no match is found.
        """
        available_streams = streams.order_by('resolution')

        return available_streams.get_by_resolution(quality) or available_streams.first()

    @yaspin(text=colored("Downloading the audio...", "green"), color="green", spinner=Spinners.dots13)
    def get_audio_streams(self, video: YouTube) -> YouTube:
        """
        Function to get audio streams from a video.

        Args:
            video: The video for which audio streams are to be obtained.

        Returns:
            The first audio stream found in the video.
        """
        return video.streams.filter(only_audio=True).order_by('mime_type').first()

    @yaspin(text=colored("Saving the file...", "cyan"), spinner=Spinners.smiley)
    def save_file(self, video: YouTube, path: str, filename: str) -> None:
        """
        Save the file to the specified path with the given filename.

        Args:
            video: The video to be saved.
            path: The path where the video will be saved.
            filename: The name of the file.

        Returns:
            None
        """
        video.download(output_path=path, filename=filename)

    def download_video(self):
        """
        Download a video from a given URL to a specified path.

        Args:
            url: The URL of the video.
            path: The path to save the downloaded video.
            quality: The quality of the video, default is '720p'.
            is_audio: Flag indicating if the video should be downloaded as audio, default is False.

        Returns:
            bool: True if the video is downloaded successfully, False otherwise.
        """
        video = self.search_process()

        if not video:
            return False

        self.video_id = video.video_id

        if self.is_audio:
            file = self.get_audio_streams(video)
        else:
            # shorts and videos
            streams = self.get_selected_stream(video)
            if not streams:
                error_console.print("❗ Cancel the download...")
                return
            file = self.get_video_streams(video, self.quality, streams)

        if not file:
            error_console.print(
                "Something went wrong while downloading the video.")
            return False

        # Generate filename with title, quality, and file extension
        filename = self.generate_filename(file, self.video_id)

        # If file with the same name already exists in the path
        if is_file_exists(self.path, filename):
            filename = self.handle_existing_file(filename)
            if not filename:
                return False
        try:
            self.save_file(file, self.path, filename)

        except Exception as error:
            error_console.print(f"Error: {error}")
            return False

        return True

    def search_process(self) -> YouTube:
        try:
            video = self.video_search(self.url)
        except Exception as error:
            error_console.print(f"Error: {error}")
            return False

        # If video is not found
        if not video:
            error_console.print(
                f"No {'audio' if self.is_audio else 'video'} stream available for the video.")
            return False

        console.print(f"Title: {video.title}\n", style="info")
        return video

    def get_selected_stream(self, video):
        """
        Get the selected video stream based on user preference.

        Returns:
            YouTube: The selected video stream.
        """
        resolutions, streams = self.get_available_resolutions(video)
        self.quality = ask_resolution(resolutions)

        return [] if self.quality.startswith("cancel") else streams

    def generate_filename(self, video, video_id):
        """
        Generate a filename for the downloaded video.

        Returns:
            str: The generated filename.
        """
        quality = 'audio' if self.is_audio else video.resolution
        title = sanitize_filename(video.title)
        return f"{title} - {quality}_-_{video_id}.{'mp3' if self.is_audio else video.mime_type.split('/')[1]}"

    def handle_existing_file(self, filename):
        """
        Handle the case where a file with the same name already exists.

        Returns:
            str: The user's choice.
        """
        choice = ask_rename_file(filename).lower()
        if choice.startswith('rename'):
            filename = self.prompt_new_filename(filename)
            if not filename:
                error_console.print("Invalid filename")
                return False
            return filename
        elif choice.startswith('cancel'):
            console.print("Download canceled", style="info")
            return False
        return True

    def prompt_new_filename(self, filename):
        """
        Prompt the user for a new filename.

        Returns:
            str: The new filename.
        """
        text = colored(filename, 'yellow')
        new_name = input(f"Rename {text} to: ")
        return rename_file(filename, new_name)


def download(url: str, path: str, is_audio: bool) -> None:
    """
    Downloads the YouTube video based on the provided parameters.

    Args:
        url (str): The URL of the YouTube video.
        path (str): The path to save the video.
        quality_choice (str): The chosen quality for the video.
        is_audio (bool): Whether the download is for audio.

    Returns:
        None
    """
    downloader = Downloader(
        url=url, path=path, is_audio=is_audio,
    )
    downloader.download_video()
    del downloader



import pytest
from pyutube.utils import is_youtube_link


@pytest.mark.parametrize("link, expected_result", [
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ",
     (True, "video")),
    ("https://youtu.be/dQw4w9WgXcQ", (True, "video")),
    # ("https://www.youtube.com/playlist?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
    #  (True, "playlist")),
    ("https://www.youtube.com/shorts/dQw4w9WgXcQ",
     (True, "shorts")),


    ("", (False, "unknown")),
    ("https://www.youtube.com/", (False, "unknown")),
    ("https://www.youtube.com/watch?v=", (False, "unknown")),
    ("https://www.youtu.be/", (False, "unknown")),
    ("https://www.youtube.com/playlist?list=",
     (False, "unknown")),
    ("https://www.youtube.com/shorts/", (False, "unknown")),


    ("https://www.google.com", (False, "unknown")),
    ("https://www.youtube.com/watch?list=dQw4w9WgXcQ",
     (False, "unknown")),
    ("https://www.youtu.be/dQw4w9WgXcQ?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
     (True, "video")),
    ("https://www.youtube.com/shorts/dQw4w9WgXcQ?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
     (True, "shorts")),
])
def test_is_youtube_link_error_cases(link, expected_result):
    # Act
    result = is_youtube_link(link)

    # Assert
    assert result == expected_result



from setuptools import setup, find_packages


# Read the README file to use as the long description
with open("README.md", "r") as f:
    description = f.read()

# Setup configuration
setup(
    # Name of the package
    name="pyutube",

    # Version of the package
    version="1.1.4",

    # Required dependencies
    install_requires=[
        "pytube",
        "inquirer",
        "yaspin",
        "typer",
        "requests",
        "rich",
        "inquirer",
        "termcolor"
    ],

    # Classifiers to categorize the package
    classifiers=[
        "License :: OSI Approved :: MIT License",
    ],

    # Entry points for console scripts
    entry_points={
        "console_scripts": [
            "pyutube=pyutube:cli.app",  # Command to run the application
        ],
    },

    # Python version requirement
    python_requires=">=3.6",

    # Long description of the package
    long_description=description,

    # Description content type
    long_description_content_type="text/markdown",

    # Project URLs
    project_urls={
        "Homepage": "https://github.com/Hetari/pyutube",
    }
)



import re
import inquirer
import requests
import os
from yaspin import yaspin
from yaspin.spinners import Spinners
from rich.console import Console
from rich.theme import Theme


ABORTED_PREFIX = "aborted"
CANCEL_PREFIX = "cancel"

# Set up the console
custom_theme = Theme({
    "info": "#64b0f2",
    "warning": "color(3)",
    "danger": "red",
    "success": "green",
})
console = Console(theme=custom_theme)
error_console = Console(stderr=True, style="red")


def clear() -> None:
    """
    Function to clear the console screen, it can be used for any operating system

    Args:
        This function does not take any parameters.

    Returns:
        It does not return anything (None).
    """
    # For Windows
    if os.name == "nt":
        os.system("cls")
    else:
        # For Unix/Linux/MacOS
        os.system("clear")


@yaspin(text="Checking internet connection", color="blue", spinner=Spinners.earth)
def is_internet_available() -> bool:
    """
    Checks if internet connection is available by making a simple request to http://www.google.com with a timeout of 5 seconds.

    Returns:
        bool: the request status (True if available, False if not).
    """
    try:
        requests.get("https://www.google.com", timeout=5)
        return True
    except requests.ConnectionError:
        return False
    except requests.ReadTimeout:
        return False
    except requests.HTTPError:
        return False
    except requests.TooManyRedirects:
        return False
    except requests.RequestException:
        return False
    except Exception as error:
        error_console.print(f"Error: {error}")
        return False


def is_youtube_link(link: str) -> tuple[bool, str]:
    """
    Check if the given link is a YouTube video, playlist, or shorts link.

    Args:
        link (str): The link to be checked.

    Returns:
        tuple[bool, str]: True if the link is a YouTube video, playlist, or shorts link,
        False otherwise. The second item of the tuple indicates the type of link found:
        'video', 'playlist', or 'shorts'.
    """
    is_video = is_youtube_video(link)
    is_short = is_youtube_shorts(link)

    return (is_video, "video") if is_video else (is_short, "short") if is_short else (False, "unknown")


def is_youtube_shorts(link: str) -> bool:
    """
    Check if the given link is a YouTube shorts link.

    Args:
        link: The link to be checked.

    Returns:
        bool: True if the link is a YouTube shorts link, False otherwise.
    """
    # shorts_pattern = r"(?:https?:\/\/)?(?:www\.)?youtube\.com\/shorts\/([a-zA-Z0-9_-]+)"
    shorts_pattern = r"(?:https?:\/\/)?(?:www\.)?(?:youtube\.com\/(?:[^\/\n\s]+\/\S+\/|shorts\/|watch\?.*?v=))(?:(?:[^\/\n\s]+\/)?)([a-zA-Z0-9_-]+)"
    shorts_match = re.match(shorts_pattern, link)
    return bool(shorts_match)


def is_youtube_video(link: str) -> bool:
    """
    Check if the given link is a YouTube video.

    Args:
        link: The link to be checked.

    Returns:
        bool: True if the link is a YouTube video, False otherwise.
    """
    # video_pattern = re.compile(
    #     r'(?:https?://)?(?:www\.)?(?:youtube\.com/(?:(?:watch\?v=)|(?:embed/))|youtu\.be/)([a-zA-Z0-9_-]{11})')

    video_pattern = re.compile(
        r'(?:https?://)?(?:www\.)?(?:youtube\.com/(?:(?:watch\?v=)|(?:embed/))|youtu\.be/|youtube.com/share\?v=)([a-zA-Z0-9_-]{11})')

    return bool(video_pattern.match(link))


def is_youtube_video_id(video_id: str) -> bool:
    """
    Check if the given string is a valid YouTube video ID.

    Args:
        video_id: The string to be checked.

    Returns:
        bool: True if the string is a valid YouTube video ID, False otherwise.
    """
    return len(video_id) == 11 and all(c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-" for c in video_id)


def file_type() -> str:
    """
    Prompts the user to choose a file type for download and returns the chosen file type as a string.

    Args:
        None

    Returns:
        str: The chosen file type as a string.
    """
    # make the console font to red
    questions = [
        inquirer.List(
            "file_type",
            message="Choose the file type you want to download",
            choices=['Audio', 'Video', "Cancel the download"],
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["file_type"]

    # TypeError: 'NoneType' object is not subscriptable

    except TypeError as error:
        return "Aborted"

    except Exception as error:
        error_console.print(f"Error: {error}")

    return answer


def ask_resolution(resolutions: set) -> str:
    """
    Prompts the user to choose a resolution for download and returns the chosen resolution as a string.

    Args:
        resolutions (set): The set of available resolutions.

    Returns:
        str: The chosen resolution as a string.
    """
    resolution_choices = list(resolutions) + ["Cancel the download"]
    questions = [
        inquirer.List(
            "resolution",
            message="Choose the resolution you want to download",
            choices=resolution_choices,
        ),
    ]

    try:
        answer = inquirer.prompt(questions)["resolution"]

    except TypeError as error:
        return "Aborted"

    except Exception as error:
        error_console.print(f"Error: {error}")

    return answer.lower()


def ask_rename_file(filename: str) -> str:
    """
    Function to ask the user whether to rename, overwrite, or cancel the file operation.

    Args:
        filename (str): The name of the file.

    Returns:
        str: The user's choice to rename, overwrite, or cancel the file operation.
    """
    console.print(
        f"'{filename}' is already exists, do you want to:", style="info")
    questions = [
        inquirer.List(
            "rename",
            message="Do you want to",
            choices=['Rename it', 'Overwrite it', "Cancel"],
        ),
    ]
    return inquirer.prompt(questions)["rename"]


def sanitize_filename(filename: str) -> str:
    """
    Removes characters that are not allowed in filenames.

    Args:
        filename: The filename to be sanitized.

    Returns:
        str: The sanitized filename.
    """
    return re.sub(r'[<>:"/\\|?*]', '_', filename)


def rename_file(filename: str, new_filename: str) -> str:
    """
    Rename a file to a new filename.

    If the new filename does not have the same file extension as the original
    filename, the extension of the original filename will be appended to the
    new filename.

    Args:
        filename: The original filename.
        new_filename: The new filename to rename to.

    Returns:
        str: The new filename after the renaming operation.
    """
    try:
        if not new_filename.endswith(os.path.splitext(filename)[1]):
            new_filename += f".{filename.split('.')[-1]}"
    except IndexError as error:
        error_console.print(f"Error: {error}")
    return new_filename


def is_file_exists(path: str, filename: str) -> bool:
    """
    Check if a file exists at the specified path and filename.

    Args:
        path: The path where the file is located.
        filename: The name of the file to check.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    return os.path.isfile(os.path.join(path, filename))


# main utils
def check_internet_connection() -> bool:
    """
    Checks if an internet connection is available.

    Returns:
        bool: True if internet connection is available, False otherwise.
    """
    if not is_internet_available():
        error_console.print("❗ No internet connection")
        return False
    else:
        console.print("✅ There is internet connection", style="info")
        console.print()
        return True


def validate_link(url: str) -> tuple[bool, str]:
    """
    Validates the given YouTube video URL.

    Args:
        url (str): The URL of the YouTube video.

    Returns:
        Tuple[bool, str]: A tuple containing a boolean indicating if the link is valid
        and a string indicating the type of the link (video or short).
    """
    is_valid_link, link_type = is_youtube_link(url)
    if not is_valid_link:
        console.print("❌ Invalid link", style="danger")

    return is_valid_link, link_type.lower()


def handle_video_link(url: str, path: str) -> None:
    """
    Handles video link scenario.

    Args:
        url (str): The URL of the YouTube video.
        path (str): The path to save the video.

    Returns:
        None
    """
    file_type_choice = file_type().lower()
    is_audio = file_type_choice.startswith("audio")

    if file_type_choice.startswith(CANCEL_PREFIX):
        error_console.print("❗ Cancel the download...")
        return
    elif file_type_choice.startswith(ABORTED_PREFIX):
        return

    return url, path, is_audio
